import tensorflow as tf


class HyperparameterScheduler(tf.keras.callbacks.Callback):
    """Generic hyperparameter scheduler.

    # Arguments
    schedule: a function that takes an epoch index as input
        (integer, indexed from 0) and returns a new hyperparameter as output.
    hyperparameter: str. the name of the hyperparameter to be scheduled.
    verbose: int. 0: quiet, 1: update messages.
    """

    def __init__(self, schedule, hyperparameter, verbose=0):
        super(HyperparameterScheduler, self).__init__()
        self.schedule = schedule
        self.hyperparameter = hyperparameter
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        if not hasattr(self.model.optimizer, self.hyperparameter):
            raise ValueError(
                f'Optimizer must have a "{self.hyperparameter}" attribute.'
            )

        hp = getattr(self.model.optimizer, self.hyperparameter)
        try:  # new API
            hyperparameter_val = tf.keras.backend.get_value(hp)
            hyperparameter_val = self.schedule(epoch, hyperparameter_val)
        except TypeError:  # Support for old API for backward compatibility
            hyperparameter_val = self.schedule(epoch)

        tf.keras.backend.set_value(hp, hyperparameter_val)

        if self.verbose > 0:
            print(
                f"Epoch {epoch + 1}: {self.hyperparameter} changning to {tf.keras.backend.get_value(hp)}."
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        hp = getattr(self.model.optimizer, self.hyperparameter)
        logs[self.hyperparameter] = tf.keras.backend.get_value(hp)



import tensorflow as tf


def test_eager_and_graph_mode_fixture(eager_and_graph_mode):
    if eager_and_graph_mode == "eager":
        assert tf.executing_eagerly()
    else:
        assert not tf.executing_eagerly()


def test_eager_mode_fixture(eager_mode):
    assert tf.executing_eagerly()


def test_graph_mode_fixture(graph_mode):
    assert not tf.executing_eagerly()



import itertools
from dataclasses import dataclass

import numpy as np
import tensorflow.keras.layers as keras_layers
from terminaltables import AsciiTable

import larq.layers as lq_layers

__all__ = ["summary"]

op_count_supported_layer_types = (
    lq_layers.QuantConv2D,
    lq_layers.QuantSeparableConv2D,
    lq_layers.QuantDepthwiseConv2D,
    lq_layers.QuantDense,
    keras_layers.Conv2D,
    keras_layers.SeparableConv2D,
    keras_layers.DepthwiseConv2D,
    keras_layers.Dense,
    keras_layers.Flatten,
    keras_layers.BatchNormalization,
    keras_layers.MaxPool2D,
    keras_layers.AveragePooling2D,
)

mac_containing_layers = (
    lq_layers.QuantConv2D,
    lq_layers.QuantSeparableConv2D,
    lq_layers.QuantDepthwiseConv2D,
    lq_layers.QuantDense,
    keras_layers.Conv2D,
    keras_layers.SeparableConv2D,
    keras_layers.DepthwiseConv2D,
    keras_layers.Dense,
)


def _flatten(lst):
    return list(itertools.chain.from_iterable(lst))


def _bitsize_as_str(bitsize):
    bitsize_names = {8: "byte", 8 * 1024: "kB"}

    try:
        return bitsize_names[bitsize]
    except KeyError:
        raise NotImplementedError()


def _number_as_readable_str(num):
    # The initial rounding here is necessary so that e.g. `999000` gets
    # formatted as `1.000 M` rather than `1000 k`
    num = float(f"{num:.3g}")

    # For numbers less than 1000, output them directly, stripping any trailing
    # zeros and decimal places.
    if num < 1000:
        return str(num).rstrip("0").rstrip(".")

    # For numbers that are at least 1000 trillion (1 quadrillion) format with
    # scientific notation (3 s.f. = 2 d.p. in scientific notation).
    if num >= 1e15:
        return f"{num:#.2E}"

    # Count the magnitude.
    magnitude = 0
    while abs(num) >= 1000 and magnitude < 4:
        magnitude += 1
        num /= 1000.0

    # ':#.3g' formats the number with 3 significant figures, without stripping
    # trailing zeros.
    num = f"{num:#.3g}".rstrip(".")
    unit = ["", " k", " M", " B", " T"][magnitude]
    return num + unit


def _format_table_entry(x, units=1):
    try:
        assert not np.isnan(x)
        if type(x) == str or x == 0 or units == 1:
            return x
        return x / units
    except Exception:
        return "?"


def _get_output_shape(layer):
    try:
        return tuple(dim if dim else -1 for dim in layer.output_shape)
    except AttributeError:
        return "multiple"
    except RuntimeError:  # output_shape unknown in Eager mode.
        return "?"


class WeightProfile:
    def __init__(self, weight, bitwidth=32, trainable=True):
        self._weight = weight
        self.bitwidth = bitwidth
        self.trainable = trainable

    @property
    def count(self):
        return int(np.prod(self._weight.shape.as_list()))

    @property
    def memory(self):
        return self.bitwidth * self.count

    @property
    def fp_equivalent_memory(self):
        return 32 * self.count

    def is_bias(self):
        return "bias" in self._weight.name


@dataclass
class OperationProfile:
    n: int
    precision: int
    op_type: str


class LayerProfile:
    def __init__(self, layer):
        self._layer = layer
        self.weight_profiles = [
            WeightProfile(
                weight,
                self._get_bitwidth(weight),
                trainable=any(weight is w for w in layer.trainable_weights),
            )
            for weight in layer.weights
        ]

        self.op_profiles = []

        if isinstance(layer, mac_containing_layers):
            for p in self.weight_profiles:
                if not p.is_bias():
                    self.op_profiles.append(
                        OperationProfile(
                            n=p.count * self.output_pixels,
                            precision=max(self.input_precision(32), p.bitwidth),
                            op_type="mac",
                        )
                    )

    @property
    def memory(self):
        return sum(p.memory for p in self.weight_profiles)

    @property
    def fp_equivalent_memory(self):
        return sum(p.fp_equivalent_memory for p in self.weight_profiles)

    def weight_count(self, bitwidth=None, trainable=None):
        count = 0
        for p in self.weight_profiles:
            if (bitwidth is None or p.bitwidth == bitwidth) and (
                trainable is None or p.trainable == trainable
            ):
                count += p.count
        return count

    def op_count(self, op_type=None, precision=None, escape="?"):
        if op_type != "mac":
            raise ValueError("Currently only counting of MAC-operations is supported.")

        if isinstance(self._layer, op_count_supported_layer_types):
            count = 0
            for op in self.op_profiles:
                if (precision is None or op.precision == precision) and (
                    op_type is None or op.op_type == op_type
                ):
                    count += op.n
            return count
        else:
            return escape

    def input_precision(self, default="-"):
        try:
            return self._layer.input_quantizer.precision
        except AttributeError:
            return default

    @property
    def output_shape(self):
        try:
            return tuple(dim if dim else -1 for dim in self._layer.output_shape)
        except AttributeError:
            return "multiple"
        except RuntimeError:  # output_shape unknown in Eager mode.
            return "?"

    @property
    def output_pixels(self):
        """Number of pixels for a single feature map (1 for fully connected layers)."""
        if len(self.output_shape) == 4:
            return int(np.prod(self.output_shape[1:3]))
        elif len(self.output_shape) == 2:
            return 1
        else:
            raise NotImplementedError()

    @property
    def unique_param_bidtwidths(self):
        return sorted(set([p.bitwidth for p in self.weight_profiles]))

    @property
    def unique_op_precisions(self):
        return sorted(set([op.precision for op in self.op_profiles]))

    def generate_table_row(self, table_config):
        row = [self._layer.name, self.input_precision(), self.output_shape]
        for i in table_config["param_bidtwidths"]:
            n = self.weight_count(i)
            n = _format_table_entry(n, table_config["param_units"])
            row.append(n)
        row.append(_format_table_entry(self.memory, table_config["memory_units"]))
        for i in table_config["mac_precisions"]:
            n = self.op_count("mac", i)
            n = _format_table_entry(n, table_config["mac_units"])
            row.append(n)

        return row

    def _get_bitwidth(self, weight):
        try:
            for quantizer, quantized_weight in zip(
                self._layer.quantizers, self._layer.quantized_latent_weights
            ):
                if quantized_weight is weight:
                    return quantizer.precision
        except AttributeError:
            pass
        return 32


class ModelProfile(LayerProfile):
    def __init__(self, model):
        self.layer_profiles = [LayerProfile(l) for l in model.layers]

    @property
    def memory(self):
        return sum(l.memory for l in self.layer_profiles)

    @property
    def fp_equivalent_memory(self):
        return sum(l.fp_equivalent_memory for l in self.layer_profiles)

    def weight_count(self, bitwidth=None, trainable=None):
        return sum(l.weight_count(bitwidth, trainable) for l in self.layer_profiles)

    def op_count(self, op_type=None, bitwidth=None):
        return sum(l.op_count(op_type, bitwidth, 0) for l in self.layer_profiles)

    @property
    def unique_param_bidtwidths(self):
        return sorted(
            set(_flatten(l.unique_param_bidtwidths for l in self.layer_profiles))
        )

    @property
    def unique_op_precisions(self):
        return sorted(
            set(_flatten(l.unique_op_precisions for l in self.layer_profiles))
        )

    def _generate_table_header(self, table_config):
        return [
            "Layer",
            "Input prec.\n(bit)",
            "Outputs",
            *(
                f"# {i}-bit\nx {table_config['param_units']}"
                for i in table_config["param_bidtwidths"]
            ),
            f"Memory\n({_bitsize_as_str(table_config['memory_units'])})",
            *(f"{i}-bit MACs" for i in table_config["mac_precisions"]),
        ]

    def _generate_table_total(self, table_config):
        row = ["Total", "", ""]
        for i in table_config["param_bidtwidths"]:
            row.append(
                _format_table_entry(self.weight_count(i), table_config["param_units"])
            )
        row.append(_format_table_entry(self.memory, table_config["memory_units"]))
        for i in table_config["mac_precisions"]:
            row.append(
                _format_table_entry(self.op_count("mac", i), table_config["mac_units"])
            )
        return row

    def generate_table(self, include_macs=True):
        table_config = {
            "param_bidtwidths": self.unique_param_bidtwidths,
            "mac_precisions": self.unique_op_precisions if include_macs else [],
            "param_units": 1,
            "memory_units": 8 * 1024,
            "mac_units": 1,
        }

        table = []

        table.append(self._generate_table_header(table_config))

        for lp in self.layer_profiles:
            table.append(lp.generate_table_row(table_config))

        table.append(self._generate_table_total(table_config))

        return table

    def generate_summary(self, include_macs=True):
        summary = [
            ["Total params", _number_as_readable_str(self.weight_count())],
            [
                "Trainable params",
                _number_as_readable_str(self.weight_count(trainable=True)),
            ],
            [
                "Non-trainable params",
                _number_as_readable_str(self.weight_count(trainable=False)),
            ],
            ["Model size:", f"{self.memory / (8*1024*1024):.2f} MB"],
            [
                "Float-32 Equivalent",
                f"{self.fp_equivalent_memory / (8*1024*1024):.2f} MB",
            ],
            [
                "Compression Ratio of Memory",
                self.memory / max(1e-8, self.fp_equivalent_memory),
            ],
        ]

        if include_macs:
            binarization_ratio = self.op_count("mac", 1) / max(
                1, self.op_count(op_type="mac")
            )
            ternarization_ratio = self.op_count("mac", 2) / max(
                1, self.op_count(op_type="mac")
            )
            summary.append(
                [
                    "Number of MACs",
                    _number_as_readable_str(self.op_count(op_type="mac")),
                ]
            )
            if binarization_ratio > 0:
                summary.append(
                    ["Ratio of MACs that are binarized", f"{binarization_ratio:.4f}"]
                )
            if ternarization_ratio > 0:
                summary.append(
                    ["Ratio of MACs that are ternarized", f"{ternarization_ratio:.4f}"]
                )

        return summary


def sanitize_table(table_data):
    return [[f"{v:.2f}" if type(v) == float else v for v in row] for row in table_data]


class LayersTable(AsciiTable):
    def __init__(self, table_data, title=None):
        super().__init__(sanitize_table(table_data), title=title)
        self.inner_column_border = False
        self.justify_columns = {
            i: "left" if i == 0 else "right" for i in range(len(table_data[0]))
        }
        self.inner_footing_row_border = True
        self.inner_heading_row_border = True


class SummaryTable(AsciiTable):
    def __init__(self, table_data, title=None):
        super().__init__(sanitize_table(table_data), title=title)
        self.inner_column_border = False
        self.inner_heading_row_border = False


def summary(model, print_fn=None, include_macs=True):
    """Prints a string summary of the network.

    The summary includes the following information per layer:

    - input precision,
    - output dimension,
    - weight count (broken down by bidtwidth),
    - memory footprint in kilobytes (`8*1024` 1-bit weights = 1 kB),
    - number of multiply-accumulate (MAC) operations broken down by precision (*optional & expermental*).

    A single MAC operation contains both a multiplication and an addition. The precision
    of a MAC operation is defined as the maximum bitwidth of its inputs.

    Additionally, the following overall statistics for the model are supplied:

    - total number of weights,
    - total number of trainable weights,
    - total number of non-trainable weights,
    - model size,
    - float-32 equivalent size: memory footprint if all weights were 32 bit,
    - compression ratio achieved by quantizing weights,
    - total number of MAC operations,
    - ratio of MAC operations that is binarized and can be accelated with XNOR-gates.

    # Arguments
    model: `tf.keras` model instance.
    print_fn: Print function to use. Defaults to `print`. You can set it to a custom
        function in order to capture the string summary.
    include_macs: whether or not to include the number of MAC-operations in the summary.

    # Raises
    ValueError: if called before the model is built.
    """

    if not model.built:
        raise ValueError(
            "This model has not yet been built. Build the model first by calling "
            "`model.build()` or calling `model.fit()` with some data, or specify an "
            "`input_shape` argument in the first layer(s) for automatic build."
        )

    if print_fn is None:
        print_fn = print

    model_profile = ModelProfile(model)
    print_fn(
        LayersTable(model_profile.generate_table(), title=f"{model.name} stats").table
    )
    print_fn(
        SummaryTable(
            model_profile.generate_summary(include_macs), title=f"{model.name} summary"
        ).table
    )



import pytest
import tensorflow as tf

from larq import metrics


def test_scope():
    assert metrics.get_training_metrics() == set()
    with metrics.scope(["flip_ratio"]):
        assert metrics.get_training_metrics() == {"flip_ratio"}
    assert metrics.get_training_metrics() == set()
    with pytest.raises(ValueError, match=r".*unknown_metric.*"):
        with metrics.scope(["flip_ratio", "unknown_metric"]):
            pass


def test_config():
    mcv = metrics.FlipRatio(
        values_shape=[3, 3], values_dtype="int16", name="mcv", dtype=tf.float16
    )
    assert mcv.name == "mcv"
    assert mcv.stateful
    assert mcv.dtype == tf.float16
    assert mcv.values_dtype == tf.int16
    assert mcv.values_shape == [3, 3]

    mcv2 = metrics.FlipRatio.from_config(mcv.get_config())
    assert mcv2.name == "mcv"
    assert mcv2.stateful
    assert mcv2.dtype == tf.float16
    assert mcv2.values_dtype == tf.int16
    assert mcv2.values_shape == [3, 3]


def test_metric(eager_mode):
    mcv = metrics.FlipRatio([2])

    assert 0 == mcv.result().numpy()
    assert 0 == mcv.total.numpy()
    assert 0 == mcv.count.numpy()

    mcv.update_state([1, 1])
    assert all([1, 1] == mcv._previous_values.numpy())
    assert 0 == mcv.total.numpy()
    assert 1 == mcv.count.numpy()
    assert 0 == mcv.result().numpy()

    mcv.update_state([2, 2])
    assert all([2, 2] == mcv._previous_values.numpy())
    assert 1 == mcv.total.numpy()
    assert 2 == mcv.count.numpy()
    assert 1 == mcv.result().numpy()

    mcv.update_state([1, 2])
    assert all([1, 2] == mcv._previous_values.numpy())
    assert 1.5 == mcv.total.numpy()
    assert 3 == mcv.count.numpy()
    assert 1.5 / 2 == mcv.result().numpy()


def test_metric_in_graph_mode(graph_mode):
    mcv = metrics.FlipRatio([2])

    new_state = tf.compat.v1.placeholder(dtype=tf.float32, shape=[2])
    update_state_op = mcv.update_state(new_state)
    metric_value = mcv.result()

    with tf.compat.v1.Session() as sess:
        sess.run(tf.compat.v1.variables_initializer(mcv.variables))

        sess.run(update_state_op, feed_dict={new_state: [1, 1]})
        sess.run(update_state_op, feed_dict={new_state: [2, 2]})
        sess.run(update_state_op, feed_dict={new_state: [1, 2]})

        previous, total, count, result = sess.run(
            [mcv._previous_values, mcv.total, mcv.count, metric_value]
        )

    assert all([1, 2] == previous)
    assert 1.5 == total
    assert 3 == count
    assert 1.5 / 2 == result



import logging

import tensorflow as tf

from larq import metrics as lq_metrics
from larq import quantizers, utils

log = logging.getLogger(__name__)


# TODO: find a good way remove duplication between QuantizerBase, QuantizerDepthwiseBase and QuantizerSeparableBase


class QuantizerBase(tf.keras.layers.Layer):
    """Base class for defining quantized layers

    `input_quantizer` and `kernel_quantizer` are the element-wise quantization
    functions to use. If both quantization functions are `None` this layer is
    equivalent to `Layer`.
    """

    def __init__(
        self, *args, input_quantizer=None, kernel_quantizer=None, metrics=None, **kwargs
    ):
        self.input_quantizer = quantizers.get(input_quantizer)
        self.kernel_quantizer = quantizers.get(kernel_quantizer)
        self.quantized_latent_weights = []
        self.quantizers = []
        self._custom_metrics = (
            metrics if metrics is not None else lq_metrics.get_training_metrics()
        )

        super().__init__(*args, **kwargs)
        if kernel_quantizer and not self.kernel_constraint:
            log.warning(
                "Using a weight quantizer without setting `kernel_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def build(self, input_shape):
        super().build(input_shape)
        if self.kernel_quantizer:
            self.quantized_latent_weights.append(self.kernel)
            self.quantizers.append(self.kernel_quantizer)
            if "flip_ratio" in self._custom_metrics:
                self.flip_ratio = lq_metrics.FlipRatio(
                    values_shape=self.kernel.shape, name=f"flip_ratio/{self.name}"
                )

    @property
    def non_trainable_weights(self):
        weights = super().non_trainable_weights
        if hasattr(self, "flip_ratio"):
            return [
                weight
                for weight in weights
                if not any(weight is metric_w for metric_w in self.flip_ratio.weights)
            ]
        return weights

    def call(self, inputs):
        if self.input_quantizer:
            inputs = self.input_quantizer(inputs)

        with utils.quantize(self, "kernel", self.kernel_quantizer) as kernel:
            if hasattr(self, "flip_ratio"):
                self.add_metric(self.flip_ratio(kernel))
            return super().call(inputs)

    def get_config(self):
        config = {
            "input_quantizer": quantizers.serialize(self.input_quantizer),
            "kernel_quantizer": quantizers.serialize(self.kernel_quantizer),
        }
        return {**super().get_config(), **config}


class QuantizerDepthwiseBase(tf.keras.layers.Layer):
    """Base class for defining quantized layers

    `input_quantizer` and `depthwise_quantizer` are the element-wise quantization
    functions to use. If both quantization functions are `None` this layer is
    equivalent to `Layer`.
    """

    def __init__(
        self,
        *args,
        input_quantizer=None,
        depthwise_quantizer=None,
        metrics=None,
        **kwargs,
    ):
        self.input_quantizer = quantizers.get(input_quantizer)
        self.depthwise_quantizer = quantizers.get(depthwise_quantizer)
        self.quantized_latent_weights = []
        self.quantizers = []
        self._custom_metrics = (
            metrics if metrics is not None else lq_metrics.get_training_metrics()
        )

        super().__init__(*args, **kwargs)
        if depthwise_quantizer and not self.depthwise_constraint:
            log.warning(
                "Using a weight quantizer without setting `depthwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def build(self, input_shape):
        super().build(input_shape)
        if self.depthwise_quantizer:
            self.quantized_latent_weights.append(self.depthwise_kernel)
            self.quantizers.append(self.depthwise_quantizer)
            if "flip_ratio" in self._custom_metrics:
                self.flip_ratio = lq_metrics.FlipRatio(
                    values_shape=self.depthwise_kernel.shape,
                    name=f"flip_ratio/{self.name}",
                )

    @property
    def non_trainable_weights(self):
        weights = super().non_trainable_weights
        if hasattr(self, "flip_ratio"):
            return [
                weight
                for weight in weights
                if not any(weight is metric_w for metric_w in self.flip_ratio.weights)
            ]
        return weights

    def call(self, inputs):
        if self.input_quantizer:
            inputs = self.input_quantizer(inputs)

        with utils.quantize(
            self, "depthwise_kernel", self.depthwise_quantizer
        ) as kernel:
            if hasattr(self, "flip_ratio"):
                self.add_metric(self.flip_ratio(kernel))
            return super().call(inputs)

    def get_config(self):
        config = {
            "input_quantizer": quantizers.serialize(self.input_quantizer),
            "depthwise_quantizer": quantizers.serialize(self.depthwise_quantizer),
        }
        return {**super().get_config(), **config}


class QuantizerSeparableBase(tf.keras.layers.Layer):
    """Base class for defining separable quantized layers

    `input_quantizer`, `depthwise_quantizer` and `pointwise_quantizer` are the
    element-wise quantization functions to use. If all quantization functions are `None`
    this layer is equivalent to `SeparableConv1D`. If `use_bias` is True and
    a bias initializer is provided, it adds a bias vector to the output.
    It then optionally applies an activation function to produce the final output.
    """

    def __init__(
        self,
        *args,
        input_quantizer=None,
        depthwise_quantizer=None,
        pointwise_quantizer=None,
        metrics=None,
        **kwargs,
    ):
        self.input_quantizer = quantizers.get(input_quantizer)
        self.depthwise_quantizer = quantizers.get(depthwise_quantizer)
        self.pointwise_quantizer = quantizers.get(pointwise_quantizer)
        self.quantized_latent_weights = []
        self.quantizers = []
        self._custom_metrics = (
            metrics if metrics is not None else lq_metrics.get_training_metrics()
        )

        super().__init__(*args, **kwargs)
        if depthwise_quantizer and not self.depthwise_constraint:
            log.warning(
                "Using `depthwise_quantizer` without setting `depthwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )
        if pointwise_quantizer and not self.pointwise_constraint:
            log.warning(
                "Using `pointwise_quantizer` without setting `pointwise_constraint` "
                "may result in starved weights (where the gradient is always zero)."
            )

    def build(self, input_shape):
        super().build(input_shape)
        if self.depthwise_quantizer:
            self.quantized_latent_weights.append(self.depthwise_kernel)
            self.quantizers.append(self.depthwise_quantizer)
            if "flip_ratio" in self._custom_metrics:
                self.depthwise_flip_ratio = lq_metrics.FlipRatio(
                    values_shape=self.depthwise_kernel.shape,
                    name=f"flip_ratio/{self.name}_depthwise",
                )
        if self.pointwise_quantizer:
            self.quantized_latent_weights.append(self.pointwise_kernel)
            self.quantizers.append(self.pointwise_quantizer)
            if "flip_ratio" in self._custom_metrics:
                self.pointwise_flip_ratio = lq_metrics.FlipRatio(
                    values_shape=self.pointwise_kernel.shape,
                    name=f"flip_ratio/{self.name}_pointwise",
                )

    @property
    def non_trainable_weights(self):
        weights = super().non_trainable_weights
        metrics_weights = []
        if hasattr(self, "depthwise_flip_ratio"):
            metrics_weights.extend(self.depthwise_flip_ratio.weights)
        if hasattr(self, "pointwise_flip_ratio"):
            metrics_weights.extend(self.pointwise_flip_ratio.weights)
        if metrics_weights:
            return [
                weight
                for weight in weights
                if not any(weight is metric_w for metric_w in metrics_weights)
            ]
        return weights

    def call(self, inputs):
        if self.input_quantizer:
            inputs = self.input_quantizer(inputs)

        with utils.quantize(
            self, "depthwise_kernel", self.depthwise_quantizer
        ) as depthwise_kernel, utils.quantize(
            self, "pointwise_kernel", self.pointwise_quantizer
        ) as pointwise_kernel:
            if hasattr(self, "depthwise_flip_ratio"):
                self.add_metric(self.depthwise_flip_ratio(depthwise_kernel))
            if hasattr(self, "pointwise_flip_ratio"):
                self.add_metric(self.pointwise_flip_ratio(pointwise_kernel))
            return super().call(inputs)

    def get_config(self):
        config = {
            "input_quantizer": quantizers.serialize(self.input_quantizer),
            "depthwise_quantizer": quantizers.serialize(self.depthwise_quantizer),
            "pointwise_quantizer": quantizers.serialize(self.pointwise_quantizer),
        }
        return {**super().get_config(), **config}



from contextlib import contextmanager

from tensorflow.keras.utils import get_custom_objects


def register_keras_custom_object(cls):
    """See https://github.com/tensorflow/addons/blob/master/tensorflow_addons/utils/keras_utils.py#L25"""
    get_custom_objects()[cls.__name__] = cls
    return cls


def register_alias(name):
    """A decorator to register a custom keras object under a given alias.
    !!! example
        ```python
        @utils.register_alias("degeneration")
        class Degeneration(tf.keras.metrics.Metric):
            pass
        ```
    """

    def register_func(cls):
        get_custom_objects()[name] = cls
        return cls

    return register_func


def set_precision(precision=32):
    """A decorator to set the precision of a quantizer function

    # Arguments
    precision: An integer defining the precision of the output.
    """

    def decorator(function):
        setattr(function, "precision", precision)
        return function

    return decorator


@contextmanager
def quantize(layer, kernel_name, quantizer):
    """Temporarily apply a quantizer to a kernel.

    This is needed, since we do not want to mutate existing references that might
    expect a tf.Variable instead of a tf.Tensor. This can happen in eager mode since
    overwriting a kernel would also mutate layer.trainable_variables which breaks
    gradient computation.
    """
    full_precision_kernel = getattr(layer, kernel_name)
    if quantizer is None:
        yield full_precision_kernel
    else:
        quantized_kernel = quantizer(full_precision_kernel)
        setattr(layer, kernel_name, quantized_kernel)
        yield quantized_kernel
        setattr(layer, kernel_name, full_precision_kernel)



"""A Quantizer defines the way of transforming a full precision input to a
quantized output and the pseudo-gradient method used for the backwards pass.

Quantizers can either be used through quantizer arguments that are supported
for Larq layers, such as `input_quantizer` and `kernel_quantizer`; or they
can be used similar to activations, i.e. either through an `Activation` layer,
or through the `activation` argument supported by all forward layer:

```python
import tensorflow as tf
import larq as lq
...
x = lq.layers.QuantDense(64, activation=None)(x)
x = lq.layers.QuantDense(64, input_quantizer="ste_sign")(x)
```

is equivalent to:

```python
x = lq.layers.QuantDense(64)(x)
x = tf.keras.layers.Activation("ste_sign")(x)
x = lq.layers.QuantDense(64)(x)
```

as well as:

```python
x = lq.layers.QuantDense(64, activation="ste_sign")(x)
x = lq.layers.QuantDense(64)(x)
```

We highly recommend using the first of these formulations: for the
other two formulations, intermediate layers - like batch normalization or
average pooling - and shortcut connections may result in non-binary input
to the convolutions.

Quantizers can either be referenced by string or called directly.
The following usages are equivalent:

```python
lq.layers.QuantDense(64, kernel_quantizer="ste_sign")
```
```python
lq.layers.QuantDense(64, kernel_quantizer=lq.quantizers.ste_sign)
```
```python
lq.layers.QuantDense(64, kernel_quantizer=lq.quantizers.SteSign(clip_value=1.0))
```
"""

import tensorflow as tf

from larq import math, utils

__all__ = [
    "ste_sign",
    "approx_sign",
    "magnitude_aware_sign",
    "swish_sign",
    "ste_tern",
    "ste_heaviside",
    "dorefa_quantizer",
    "SteSign",
    "MagnitudeAwareSign",
    "SwishSign",
    "SteTern",
    "SteHeaviside",
    "DoReFaQuantizer",
]


def _clipped_gradient(x, dy, clip_value):
    if clip_value is None:
        return dy

    zeros = tf.zeros_like(dy)
    mask = tf.math.less_equal(tf.math.abs(x), clip_value)
    return tf.where(mask, dy, zeros)


class QuantizerFunctionWrapper:
    """Wraps a quantizer function in a class that can be serialized.

    # Arguments
    fn: The quantizer function to wrap, with signature `fn(x, **kwargs)`.
    **kwargs: The keyword arguments that are passed on to `fn`.
    """

    def __init__(self, fn, **kwargs):
        self.fn = fn
        self.precision = getattr(fn, "precision", 32)
        self._fn_kwargs = kwargs

    def __call__(self, x):
        """Invokes the `QuantizerFunctionWrapper` instance.

        # Arguments
        x: Input tensor.

        # Returns
        Quantized tensor.
        """
        return self.fn(x, **self._fn_kwargs)

    def get_config(self):
        return {
            k: tf.keras.backend.eval(v)
            if tf.is_tensor(v) or isinstance(v, tf.Variable)
            else v
            for k, v in self._fn_kwargs.items()
        }


@utils.register_keras_custom_object
@utils.set_precision(1)
def ste_sign(x, clip_value=1.0):
    r"""Sign binarization function.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
      1 & \left|x\right| \leq \texttt{clip_value} \\\
      0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_sign
    ```

    # Arguments
    x: Input tensor.
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # Returns
    Binarized tensor.

    # References
    - [Binarized Neural Networks: Training Deep Neural Networks with Weights and
      Activations Constrained to +1 or -1](http://arxiv.org/abs/1602.02830)
    """

    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return math.sign(x), grad

    return _call(x)


def _scaled_sign(x):  # pragma: no cover
    return 1.3 * ste_sign(x)


@utils.register_keras_custom_object
@utils.set_precision(1)
def magnitude_aware_sign(x, clip_value=1.0):
    r"""Magnitude-aware sign for Bi-Real Net.

    A scaled sign function computed according to Section 3.3 in
    [Zechun Liu et al](https://arxiv.org/abs/1808.00278).

    ```plot-activation
    quantizers._scaled_sign
    ```

    # Arguments
    x: Input tensor
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # Returns
    Scaled binarized tensor (with values in \\(\\{-a, a\\}\\), where \\(a\\) is a float).

    # References
    - [Bi-Real Net: Enhancing the Performance of 1-bit CNNs With Improved
      Representational Capability and Advanced Training
      Algorithm](https://arxiv.org/abs/1808.00278)

    """
    scale_factor = tf.reduce_mean(tf.abs(x), axis=list(range(len(x.shape) - 1)))

    return tf.stop_gradient(scale_factor) * ste_sign(x, clip_value=clip_value)


@utils.register_keras_custom_object
@utils.set_precision(1)
@tf.custom_gradient
def approx_sign(x):
    r"""
    Sign binarization function.
    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the ApproxSign method.
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
      (2 - 2 \left|x\right|) & \left|x\right| \leq 1 \\\
      0 & \left|x\right| > 1
    \end{cases}
    \\]

    ```plot-activation
    quantizers.approx_sign
    ```

    # Arguments
    x: Input tensor.

    # Returns
    Binarized tensor.

    # References
    - [Bi-Real Net: Enhancing the Performance of 1-bit CNNs With Improved
      Representational Capability and Advanced
      Training Algorithm](http://arxiv.org/abs/1808.00278)
    """

    def grad(dy):
        abs_x = tf.math.abs(x)
        zeros = tf.zeros_like(dy)
        mask = tf.math.less_equal(abs_x, 1.0)
        return tf.where(mask, (1 - abs_x) * 2 * dy, zeros)

    return math.sign(x), grad


@utils.register_keras_custom_object
@utils.set_precision(1)
def swish_sign(x, beta=5.0):
    r"""Sign binarization function.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the SignSwish method.

    \\[
    \frac{\partial q_{\beta}(x)}{\partial x} = \frac{\beta\left\\{2-\beta x \tanh \left(\frac{\beta x}{2}\right)\right\\}}{1+\cosh (\beta x)}
    \\]

    ```plot-activation
    quantizers.swish_sign
    ```
    # Arguments
    x: Input tensor.
    beta: Larger values result in a closer approximation to the derivative of the sign.

    # Returns
    Binarized tensor.

    # References
    - [BNN+: Improved Binary Network Training](https://arxiv.org/abs/1812.11800)
    """

    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            b_x = beta * x
            return dy * beta * (2 - b_x * tf.tanh(b_x * 0.5)) / (1 + tf.cosh(b_x))

        return math.sign(x), grad

    return _call(x)


@utils.register_keras_custom_object
@utils.set_precision(2)
def ste_tern(x, threshold_value=0.05, ternary_weight_networks=False, clip_value=1.0):
    r"""Ternarization function.

    \\[
    q(x) = \begin{cases}
    +1 & x > \Delta \\\
    0 & |x| < \Delta \\\
     -1 & x < - \Delta
    \end{cases}
    \\]

    where \\(\Delta\\) is defined as the threshold and can be passed as an argument,
    or can be calculated as per the Ternary Weight Networks original paper, such that

    \\[
    \Delta = \frac{0.7}{n} \sum_{i=1}^{n} |W_i|
    \\]
    where we assume that \\(W_i\\) is generated from a normal distribution.

    The gradient is estimated using the Straight-Through Estimator
    (essentially the Ternarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq \texttt{clip_value} \\\
    0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_tern
    ```

    # Arguments
    x: Input tensor.
    threshold_value: The value for the threshold, \\(\Delta\\).
    ternary_weight_networks: Boolean of whether to use the
        Ternary Weight Networks threshold calculation.
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # Returns
    Ternarized tensor.

    # References
    - [Ternary Weight Networks](http://arxiv.org/abs/1605.04711)
    """

    @tf.custom_gradient
    def _call(x):
        if ternary_weight_networks:
            threshold = 0.7 * tf.reduce_sum(tf.abs(x)) / tf.cast(tf.size(x), x.dtype)
        else:
            threshold = threshold_value

        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return tf.sign(tf.sign(x + threshold) + tf.sign(x - threshold)), grad

    return _call(x)


@utils.register_keras_custom_object
@utils.set_precision(1)
def ste_heaviside(x, clip_value=1.0):
    r"""
    Binarization function with output values 0 and 1.

    \\[
    q(x) = \begin{cases}
    +1 & x > 0 \\\
    0 & x \leq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).

    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq 1 \\\
    0 & \left|x\right| > 1
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_heaviside
    ```

    # Arguments
    x: Input tensor.
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # Returns
    AND-binarized tensor.
    """

    @tf.custom_gradient
    def _call(x):
        def grad(dy):
            return _clipped_gradient(x, dy, clip_value)

        return math.heaviside(x), grad

    return _call(x)


@utils.register_keras_custom_object
@utils.set_precision(2)
def dorefa_quantizer(x, k_bit=2):
    r"""k_bit quantizer as in the DoReFa paper.

    \\[
    q(x) = \begin{cases}
    0 & x < \frac{1}{2n} \\\
    \frac{i}{n} & \frac{2i-1}{2n} < |x| < \frac{2i+1}{2n} \text{ for } i \in \\{1,n-1\\}\\\
     1 & \frac{2n-1}{2n} < x
    \end{cases}
    \\]

    where \\(n = 2^{\text{k_bit}} - 1\\). The number of bits, k_bit, needs to be passed as an argument.
    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 &  0 \leq x \leq 1 \\\
    0 & \text{else}
    \end{cases}\\]

    ```plot-activation
    quantizers.dorefa_quantizer
    ```

    # Arguments
    k_bit: number of bits for the quantization.

    # Returns
    quantized tensor

    # References
    - [DoReFa-Net: Training Low Bitwidth Convolutional Neural Networks
      with Low Bitwidth Gradients](https://arxiv.org/abs/1606.06160)
    """
    x = tf.clip_by_value(x, 0.0, 1.0)

    @tf.custom_gradient
    def _k_bit_with_identity_grad(x):
        n = 2 ** k_bit - 1
        return tf.round(x * n) / n, lambda dy: dy

    return _k_bit_with_identity_grad(x)


@utils.register_keras_custom_object
class SteSign(QuantizerFunctionWrapper):
    r"""Instantiates a serializable binary quantizer.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
      1 & \left|x\right| \leq \texttt{clip_value} \\\
      0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_sign
    ```

    # Arguments
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # References
    - [Binarized Neural Networks: Training Deep Neural Networks with Weights and
      Activations Constrained to +1 or -1](http://arxiv.org/abs/1602.02830)
    """

    def __init__(self, clip_value=1.0):
        super().__init__(ste_sign, clip_value=clip_value)


@utils.register_keras_custom_object
class SteHeaviside(QuantizerFunctionWrapper):
    r"""
    Instantiates a binarization quantizer with output values 0 and 1.
    \\[
    q(x) = \begin{cases}
    +1 & x > 0 \\\
    0 & x \leq 0
    \end{cases}
    \\]

    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).

    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq 1 \\\
    0 & \left|x\right| > 1
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_heaviside
    ```

    # Arguments
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # Returns
    AND Binarization function
    """

    def __init__(self, clip_value=1.0):
        super().__init__(ste_heaviside, clip_value=clip_value)


@utils.register_keras_custom_object
class SwishSign(QuantizerFunctionWrapper):
    r"""Sign binarization function.

    \\[
    q(x) = \begin{cases}
      -1 & x < 0 \\\
      1 & x \geq 0
    \end{cases}
    \\]

    The gradient is estimated using the SignSwish method.

    \\[
    \frac{\partial q_{\beta}(x)}{\partial x} = \frac{\beta\left\\{2-\beta x \tanh \left(\frac{\beta x}{2}\right)\right\\}}{1+\cosh (\beta x)}
    \\]

    ```plot-activation
    quantizers.swish_sign
    ```
    # Arguments
    beta: Larger values result in a closer approximation to the derivative of the sign.

    # Returns
    SwishSign quantization function

    # References
    - [BNN+: Improved Binary Network Training](https://arxiv.org/abs/1812.11800)
    """

    def __init__(self, beta=5.0):
        super().__init__(swish_sign, beta=beta)


@utils.register_keras_custom_object
class MagnitudeAwareSign(QuantizerFunctionWrapper):
    r"""Instantiates a serializable magnitude-aware sign quantizer for Bi-Real Net.

    A scaled sign function computed according to Section 3.3 in
    [Zechun Liu et al](https://arxiv.org/abs/1808.00278).

    ```plot-activation
    quantizers._scaled_sign
    ```

    # Arguments
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # References
    - [Bi-Real Net: Enhancing the Performance of 1-bit CNNs With Improved
      Representational Capability and Advanced Training
      Algorithm](https://arxiv.org/abs/1808.00278)

    """

    def __init__(self, clip_value=1.0):
        super().__init__(magnitude_aware_sign, clip_value=clip_value)


@utils.register_keras_custom_object
class SteTern(QuantizerFunctionWrapper):
    r"""Instantiates a serializable ternarization quantizer.

    \\[
    q(x) = \begin{cases}
    +1 & x > \Delta \\\
    0 & |x| < \Delta \\\
     -1 & x < - \Delta
    \end{cases}
    \\]

    where \\(\Delta\\) is defined as the threshold and can be passed as an argument,
    or can be calculated as per the Ternary Weight Networks original paper, such that

    \\[
    \Delta = \frac{0.7}{n} \sum_{i=1}^{n} |W_i|
    \\]
    where we assume that \\(W_i\\) is generated from a normal distribution.

    The gradient is estimated using the Straight-Through Estimator
    (essentially the Ternarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 & \left|x\right| \leq \texttt{clip_value} \\\
    0 & \left|x\right| > \texttt{clip_value}
    \end{cases}\\]

    ```plot-activation
    quantizers.ste_tern
    ```

    # Arguments
    threshold_value: The value for the threshold, \\(\Delta\\).
    ternary_weight_networks: Boolean of whether to use the
        Ternary Weight Networks threshold calculation.
    clip_value: Threshold for clipping gradients. If `None` gradients are not clipped.

    # References
    - [Ternary Weight Networks](http://arxiv.org/abs/1605.04711)
    """

    def __init__(
        self, threshold_value=0.05, ternary_weight_networks=False, clip_value=1.0
    ):
        super().__init__(
            ste_tern,
            threshold_value=threshold_value,
            ternary_weight_networks=ternary_weight_networks,
            clip_value=clip_value,
        )


@utils.register_keras_custom_object
class DoReFaQuantizer(QuantizerFunctionWrapper):
    r"""Instantiates a serializable k_bit quantizer as in the DoReFa paper.

    \\[
    q(x) = \begin{cases}
    0 & x < \frac{1}{2n} \\\
    \frac{i}{n} & \frac{2i-1}{2n} < |x| < \frac{2i+1}{2n} \text{ for } i \in \\{1,n-1\\}\\\
     1 & \frac{2n-1}{2n} < x
    \end{cases}
    \\]

    where \\(n = 2^{\text{k_bit}} - 1\\). The number of bits, k_bit, needs to be passed as an argument.
    The gradient is estimated using the Straight-Through Estimator
    (essentially the binarization is replaced by a clipped identity on the
    backward pass).
    \\[\frac{\partial q(x)}{\partial x} = \begin{cases}
    1 &  0 \leq x \leq 1 \\\
    0 & \text{else}
    \end{cases}\\]

    ```plot-activation
    quantizers.dorefa_quantizer
    ```

    # Arguments
    k_bit: number of bits for the quantization.

    # Returns
    Quantization function

    # References
    - [DoReFa-Net: Training Low Bitwidth Convolutional Neural Networks
      with Low Bitwidth Gradients](https://arxiv.org/abs/1606.06160)
    """

    def __init__(self, k_bit):
        super().__init__(dorefa_quantizer, k_bit=k_bit)
        self.precision = k_bit


def serialize(quantizer):
    return tf.keras.utils.serialize_keras_object(quantizer)


def deserialize(name, custom_objects=None):
    return tf.keras.utils.deserialize_keras_object(
        name,
        module_objects=globals(),
        custom_objects=custom_objects,
        printable_module_name="quantization function",
    )


def get(identifier):
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        return deserialize(identifier)
    if isinstance(identifier, str):
        return deserialize(str(identifier))
    if callable(identifier):
        return identifier
    raise ValueError(
        f"Could not interpret quantization function identifier: {identifier}"
    )



import tensorflow as tf


class LogHistory(tf.keras.callbacks.Callback):
    def on_train_begin(self, logs={}):
        self.batches = []
        self.epochs = []

    def _store_logs(self, storage, batch_or_epoch, logs={}):
        if [key for key in logs if "changed_quantization_ration" in key]:
            storage.append(batch_or_epoch)

    def on_batch_end(self, batch, logs={}):
        self._store_logs(self.batches, batch, logs)

    def on_epoch_end(self, epoch, logs={}):
        self._store_logs(self.epochs, epoch, logs)


class TestHyperparameterScheduler:
    pass



import pytest
import tensorflow as tf
from tensorflow.python.eager import context


@pytest.fixture
def eager_mode():
    """pytest fixture for running test in eager mode"""
    with context.eager_mode():
        yield


@pytest.fixture
def graph_mode():
    """pytest fixture for running test in graph mode"""
    with context.graph_mode():
        yield


@pytest.fixture(params=["eager", "graph"])
def eager_and_graph_mode(request):
    """pytest fixture for running test in eager and graph mode"""
    with getattr(context, f"{request.param}_mode")():
        yield request.param


@pytest.fixture(params=["graph", "tf_eager", "tf_keras_eager"])
def keras_should_run_eagerly(request):
    """Fixture to run in graph and two eager modes.

    The modes are:
    - Graph mode
    - TensorFlow eager and Keras eager
    - TensorFlow eager and Keras not eager

    The `tf.context` sets graph/eager mode for TensorFlow. The yield is True if Keras
    should run eagerly.
    """

    if request.param == "graph":
        if int(tf.__version__[0]) >= 2:
            pytest.skip("Skipping graph mode for TensorFlow 2+.")

        with context.graph_mode():
            yield
    else:
        with context.eager_mode():
            yield request.param == "tf_keras_eager"



"""We add metrics specific to extremely quantized networks using a `scope` rather than
through the `metrics` parameter of `model.compile()`, where most common metrics reside.
This is because, to calculate metrics like the `flip_ratio`, we need a layer's kernel or
activation and not just the `y_true` and `y_pred` that Keras passes to metrics defined
in the usual way.
"""

from contextlib import contextmanager

import numpy as np
import tensorflow as tf

from larq import utils

__all__ = ["scope", "get_training_metrics"]

_GLOBAL_TRAINING_METRICS = set()
_AVAILABLE_METRICS = {"flip_ratio"}


@contextmanager
def scope(metrics=[]):
    """A context manager to set the training metrics to be used in layers.

    !!! example
        ```python
        with larq.metrics.scope(["flip_ratio"]):
            model = tf.keras.models.Sequential(
                [larq.layers.QuantDense(3, kernel_quantizer="ste_sign", input_shape=(32,))]
            )
        model.compile(loss="mse", optimizer="sgd")
        ```

    # Arguments
    metrics: Iterable of metrics to add to layers defined inside this context.
        Currently only the `flip_ratio` metric is available.
    """
    for metric in metrics:
        if metric not in _AVAILABLE_METRICS:
            raise ValueError(
                f"Unknown training metric '{metric}'. Available metrics: {_AVAILABLE_METRICS}."
            )
    backup = _GLOBAL_TRAINING_METRICS.copy()
    _GLOBAL_TRAINING_METRICS.update(metrics)
    yield _GLOBAL_TRAINING_METRICS
    _GLOBAL_TRAINING_METRICS.clear()
    _GLOBAL_TRAINING_METRICS.update(backup)


def get_training_metrics():
    """Retrieves a live reference to the training metrics in the current scope.

    Updating and clearing training metrics using `larq.metrics.scope` is preferred,
    but `get_training_metrics` can be used to directly access them.

    !!! example
        ```python
        get_training_metrics().clear()
        get_training_metrics().add("flip_ratio")
        ```

    # Returns
    A set of training metrics in the current scope.
    """
    return _GLOBAL_TRAINING_METRICS


@utils.register_alias("flip_ratio")
@utils.register_keras_custom_object
class FlipRatio(tf.keras.metrics.Metric):
    """Computes the mean ration of changed values in a given tensor.

    !!! example
        ```python
        m = metrics.FlipRatio(values_shape=(2,))
        m.update_state((1, 1))  # result: 0
        m.update_state((2, 2))  # result: 1
        m.update_state((1, 2))  # result: 0.75
        print('Final result: ', m.result().numpy())  # Final result: 0.75
        ```

    # Arguments
    values_shape: Shape of the tensor for which to track changes.
    values_dtype: Data type of the tensor for which to track changes.
    name: Name of the metric.
    dtype: Data type of the moving mean.
    """

    def __init__(
        self, values_shape=(), values_dtype="int8", name="flip_ratio", dtype=None
    ):
        super().__init__(name=name, dtype=dtype)
        self.values_dtype = tf.as_dtype(values_dtype)
        self.values_shape = tf.TensorShape(values_shape).as_list()
        self.is_weight_metric = True
        with tf.init_scope():
            self._previous_values = self.add_weight(
                "previous_values",
                shape=values_shape,
                dtype=self.values_dtype,
                initializer=tf.keras.initializers.zeros,
                aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
            )
            self.total = self.add_weight(
                "total",
                initializer=tf.keras.initializers.zeros,
                aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
            )
            self.count = self.add_weight(
                "count",
                initializer=tf.keras.initializers.zeros,
                aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
            )
        self._size = np.prod(self.values_shape)

    def update_state(self, values, sample_weight=None):
        values = tf.cast(values, self.values_dtype)
        changed_values = tf.math.count_nonzero(tf.equal(self._previous_values, values))
        flip_ratio = 1 - (tf.cast(changed_values, self.dtype) / self._size)

        update_total_op = self.total.assign_add(flip_ratio * tf.sign(self.count))
        with tf.control_dependencies([update_total_op]):
            update_count_op = self.count.assign_add(1)
            with tf.control_dependencies([update_count_op]):
                return self._previous_values.assign(values)

    def result(self):
        return tf.compat.v1.div_no_nan(self.total, self.count - 1)

    def reset_states(self):
        tf.keras.backend.batch_set_value(
            [(v, 0) for v in self.variables if v is not self._previous_values]
        )

    def get_config(self):
        return {
            **super().get_config(),
            "values_shape": self.values_shape,
            "values_dtype": self.values_dtype.name,
        }
